/*
 *
 */

#include "../include/nokernel.h"
#include <unistd.h>
#include <dirent.h>
/*#include <sys/stat.h>*/



struct super_block g_sb;
extern struct key root_key;


/*------------------------------------------------------------------------*/

/* de->d_iname contains name */
static struct inode * get_inode_by_name (struct inode * dir, struct dentry * dentry)
{
  dentry->d_name.len = strlen (dentry->d_iname);
  dentry->d_name.name = dentry->d_iname;

  reiserfs_lookup (dir, dentry);

  return dentry->d_inode;
}


char cmdname[10];

struct inode * pwd;

static void do_create (char * cmd)
{
  struct dentry de;

  if (sscanf (cmd, "%10s %255s", cmdname, de.d_iname) != 2) {
    reiserfs_warning ("create: usage: create filename\n");
    return;
  }

  if (get_inode_by_name (pwd, &de) == 0) {
    reiserfs_create (pwd, &de, 0100644);
  } else
    reiserfs_warning ("create: file %s exists\n", de.d_name.name);

}

static void do_mkdir (char * cmd)
{
  struct dentry de;

  if (sscanf (cmd, "%10s %255s", cmdname, de.d_iname) != 2) {
    reiserfs_warning ("mkdir: usage: mkdir dirname\n");
    return;
  }

  if (get_inode_by_name (pwd, &de) == 0) {
    reiserfs_mkdir (pwd, &de, 0100644);
  } else
    reiserfs_warning ("mkdir: dir %s exists\n", de.d_name.name);
}

static void do_rmdir (char * cmd)
{
  struct inode * dir;
  struct dentry de;
  struct path path;

  if (sscanf (cmd, "%10s %255s", cmdname, de.d_iname) != 2) {
    reiserfs_warning ("rmdir: usage: rmdir dirname\n");
    return;
  }

  if (get_inode_by_name (pwd, &de) != 0) {
    reiserfs_rmdir (pwd, &de);
  } else
    reiserfs_warning ("rmdir: dir %s is not exists\n", de.d_name.name);
}


static void do_write (char * cmd)
{
  int i;
  int count;
  loff_t offset;
  struct file file;
  char * buf;
  struct dentry de;

  if (sscanf (cmd, "%10s %255s %d %d", cmdname, de.d_iname, &offset, &count) != 4) {
    reiserfs_warning ("write: usage: write filename offset count\n");
    return;
  }

  if (get_inode_by_name (pwd, &de)) {
    file.f_dentry = &de;
    file.f_flags = 0;
    /* if regular file */
    file.f_op = reiserfs_file_inode_operations.default_file_ops;
    buf = (char *)malloc (count);
    if (buf == 0)
      reiserfs_panic (&g_sb, "do_write: no memory, or function not defined");
    for (i = 0; i < count; i ++)
      buf[i] = '0' + i % 10;

    file.f_op->write (&file, buf, count, &offset);
    free (buf);
  }
}

static void do_read (char * cmd)
{
  int count;
  loff_t offset;
  struct file file;
  char * buf;
  struct dentry de;

  if (sscanf (cmd, "%10s %255s %d %d", cmdname, de.d_iname, &offset, &count) != 4) {
    reiserfs_warning ("do_read: usage: read filename offset count\n");
    return;
  }

  if (get_inode_by_name (pwd, &de)) {
    file.f_dentry = &de;
    file.f_flags = 0;
    file.f_pos = 0;
    /* if regular file */
    file.f_op = reiserfs_file_inode_operations.default_file_ops;
    buf = (char *)malloc (count);
    if (buf == 0)
      reiserfs_panic (&g_sb, "do_read: no memory, or function not defined");
    memset (buf, 0, count);

    file.f_op->read (&file, buf, count, &offset);
    free (buf);
  }
}



static void do_fcopy (char * cmd)
{
  char source[260];
  char dest[260];
  int fd_source;
  int rd,bufsize;
  loff_t offset;
  struct file file;
  char * buf;
  struct dentry de;
  struct path path;

  if (sscanf (cmd, "%10s %255s %255s", cmdname, source, dest) != 3) {
    reiserfs_warning ("fcopy: usage: fcopy source dest\n");
    return;
  }

  fd_source=open(source, O_RDONLY);
  if (fd_source == -1) 
    reiserfs_panic (&g_sb, "fcopy: source file could not open");

  bufsize = 1024;
  buf = (char *)malloc (bufsize);
  if (buf == 0)
      reiserfs_panic (&g_sb, "fcopy: no memory, or function not defined");

  strcpy(de.d_iname,dest);


  if (get_inode_by_name (pwd, &de) == 0) {
    reiserfs_create (pwd, &de, 0100644);
  } else {
    reiserfs_warning ("fcopy: file %s exists\n", de.d_name.name);
    return;
  }
  

  file.f_dentry = &de;
  file.f_flags = 0;
  /* if regular file */
  file.f_op = reiserfs_file_inode_operations.default_file_ops;
  
  offset = 0;
  while((rd = read(fd_source, buf, bufsize)) > 0)
    {
      file.f_op->write (&file, buf, rd, &offset);
/*      offset += bufsize;*/
    }
  free (buf);
  
  close(fd_source);
}


/****************************** readdir *****************************/

struct linux_dirent {
	unsigned long	d_ino;
	unsigned long	d_off;
	unsigned short	d_reclen;
	char		d_name[1];
};

struct getdents_callback {
	struct linux_dirent * current_dir;
	struct linux_dirent * previous;
	int count;
	int error;
};

#define NAME_OFFSET(de) ((int) ((de)->d_name - (char *) (de)))
#define ROUND_UP(x) (((x)+sizeof(long)-1) & ~(sizeof(long)-1))
static int filldir(void * __buf, const char * name, int namlen, off_t offset, ino_t ino)
{
  struct linux_dirent * dirent;
  struct getdents_callback * buf = (struct getdents_callback *) __buf;
  int reclen = ROUND_UP(NAME_OFFSET(dirent) + namlen + 1);

  buf->error = -EINVAL;	/* only used if we fail.. */
  if (reclen > buf->count)
    return -EINVAL;
  dirent = buf->previous;
  if (dirent)
    put_user(offset, &dirent->d_off);
  dirent = buf->current_dir;
  buf->previous = dirent;
  put_user(ino, &dirent->d_ino);
  put_user(reclen, &dirent->d_reclen);
  copy_to_user(dirent->d_name, name, namlen);
  put_user(0, dirent->d_name + namlen);

printf ("\"%s\"\n", dirent->d_name);

  ((char *) dirent) += reclen;
  buf->current_dir = dirent;
  buf->count -= reclen;
  return 0;
}


int emu_getdents (struct file * file, void * dirent, int count)
{
  struct getdents_callback buf;
  struct linux_dirent * lastdirent;
  int error;

  buf.current_dir = (struct linux_dirent *) dirent;
  buf.previous = NULL;
  buf.count = count;
  buf.error = 0;
  file->f_op->readdir (file, &buf, filldir);
  error = buf.error;
  lastdirent = buf.previous;
  if (lastdirent) {
    put_user(file->f_pos, &lastdirent->d_off);
    error = count - buf.count;
  }

  return error;
}


void do_readdir (void)
{
  struct dentry de;
  struct file file;
  struct dirent * dirent;

  de.d_inode = pwd;
  file.f_dentry = &de;
  file.f_pos = 0;
  file.f_op = reiserfs_dir_inode_operations.default_file_ops;

  dirent = (struct dirent *)malloc (sizeof (struct dirent) * 5);
  while (emu_getdents (&file, dirent, sizeof (struct dirent) * 5) != 0)
    ;
  free (dirent);
}


void do_cd (char * cmd)
{
  struct inode * dir;
  struct dentry de;
  struct path path;

  if (sscanf (cmd, "%10s %255s", cmdname, de.d_iname) != 2) {
    reiserfs_warning ("do_cd: usage: cd dirname\n");
    return;
  }
  dir = get_inode_by_name (pwd, &de);
  if (dir != 0 && S_ISDIR (dir->i_mode))
    pwd = dir;
  else
    reiserfs_warning ("do_cd: no such file or not a directory\n");
}



#include <dirent.h>

void do_dcopy (char * cmd)
{
  char name[256], * p;
  char command [256];
  DIR * d;
  struct dirent * de;
  struct stat st;

  if (sscanf (cmd, "%10s %255s", cmdname, name) != 2) {
    reiserfs_warning ("do_dcopy: usage: dcopy dirname\n");
    return;
  }
  if ((d = opendir (name)) == NULL || chdir (name) == -1) {
    printf ("%s\n", strerror (errno));
    return;
  }

  p = strrchr (name, '/');
  p ++;
  sprintf (command, "mkdir %s\n", p);
  do_mkdir (command);
  sprintf (command, "cd %s\n", name, p);
  do_cd (command);

  while ((de = readdir (d)) != NULL) {
    if (stat (de->d_name, &st) == -1) {
      printf ("%s\n", strerror (errno));
      return;
    }
    if (!S_ISREG (st.st_mode)) {
      printf ("%s skipped\n", de->d_name);
      continue;
    }
    printf ("%s/%s\n", name, de->d_name);
    sprintf (command, "fcopy %s/%s %s\n", name, de->d_name, de->d_name);
    do_fcopy (command);
  }
  
}


char buf1[1024], buf2[1024];

void do_diff (char * cmd)
{
  char name[81], orig[256];
  int fd, rd1, rd2;
  struct file file;
  struct dentry de;

  if (sscanf (cmd, "%10s %80s %255s", cmdname, de.d_iname, orig) != 3) {
    reiserfs_warning ("diff: usage: diff filename sourcefilename\n");
    return;
  }

  fd = open (orig, O_RDONLY);
  if (fd == -1) {
    printf ("%s\n", strerror (errno));
    return;
  }

  /* open file on reiserfs */
  if (get_inode_by_name (pwd, &de)) {
    file.f_dentry = &de;
    file.f_flags = 0;
    file.f_pos = 0;
    /* if regular file */
    file.f_op = reiserfs_file_inode_operations.default_file_ops;
  } else {
    printf ("No such file or directory\n");
    return;
  }
  while ((rd1 = read (fd, buf1, 1024)) > 0) {
    rd2 = file.f_op->read (&file, buf2, 1024, &file.f_pos);
    if (rd1 != rd2) {
      printf ("Read error 1\n");
      return;
    }
    if (memcmp (buf1, buf2, rd1)) {
      printf ("Read error 2\n");
      return;
    }
  }
}

void do_delete (char * cmd)
{
  struct dentry de;

  if (sscanf (cmd, "%10s %255s", cmdname, de.d_iname) != 2) {
    reiserfs_warning ("delete: usage: delete filename\n");
    return;
  }
  if (get_inode_by_name (pwd, &de) == 0) {
    reiserfs_warning ("delete: file %s does not exist\n", de.d_name.name);
  } else {
    reiserfs_unlink (pwd, &de);
    reiserfs_delete_inode (de.d_inode);
  }

}

void do_help (void)
{
  printf (" p      <dir>                         - print the disk tree\n");
  printf (" mkdir  <dir_name>                    - create the directory\n");
  printf (" rmdir  <dir_name>                    - remove the directory\n");
  printf (" dcopy  <source> <dest>               - not ready. copy directories: source(ext2) to dest(reiserfs)\n");
  printf (" fcopy  <source> <dest>               - copy files: source(ext2) to dest(reiserfs)\n");
  printf (" create <file_name>                   - create the file\n");
  printf (" write  <file_name> <offset> <count>  - wite to the file\n");
  printf (" q                                    - quit emu\n");
}


void release_bitmaps (struct super_block * s)
{
  int i;

  for (i = 0; i < SB_BMAP_NR (s); i ++) {
    brelse (SB_AP_CAUTIOUS_BITMAP (s)[i]);
    brelse (SB_AP_TRUE_BITMAP (s)[i]);
  }

  kfree (SB_AP_TRUE_BITMAP (s));
  kfree (SB_AP_CAUTIOUS_BITMAP (s));
}

void init_pwd (void)
{
  struct path path;
  
  init_path (&path);
  
  pwd = reiserfs_iget (&g_sb, &path, &root_key);
}


int main (int argc, char * argv [])
{
  char cmd[81];
  char * file_name;
  int dev;

  printf ("\n<-----------REISERSF LAUNCHER, version 0.91a, 1999----------->\n");

#if 0
  if (0) {
    /* ???????? */
    getblk(0,0,0);
    iput(0);
    reiserfs_read_super (0,0,0);
  }
#endif

  if(argc < 2) 
    reiserfs_panic (0, "Usage: emu <device>\n");

  file_name = argv[1];

  /* open_device will die if it could not open device */
  dev = open (file_name, O_RDWR);
  if (dev == -1)
    reiserfs_panic (0, "emu: can not open '%s': %s", file_name, strerror (errno));

  init_buffer_cache ();

  read_super_block (dev, &g_sb);
  read_bitmaps (&g_sb);
 
  /* check whether device contains mounted tree file system */
  if (is_mounted (file_name))
    reiserfs_panic (0, "emu: '%s' contains a not mounted file system\n", file_name);

  init_pwd ();


  while (1) {
    printf ("Enter command: >");
    fgets (cmd, 80, stdin);
    if (strncasecmp (cmd, "create", 6) == 0)
      do_create (cmd);
    if (strncasecmp (cmd, "delete", 6) == 0)
      do_delete (cmd);
    else if (strncasecmp (cmd, "write", 5) == 0)
      do_write (cmd);    
    else if (strncasecmp (cmd, "read", 4) == 0)
      do_read (cmd);    
    else if (strncasecmp (cmd, "p", 1) == 0)
      ;/*do_print (cmd);    */
    else if (strncasecmp (cmd, "mkdir", 5) == 0)
      do_mkdir (cmd);    
    else if (strncasecmp (cmd, "rmdir", 5) == 0)
      do_rmdir (cmd);    
    else if (strncasecmp (cmd, "dcopy", 5) == 0)
      do_dcopy (cmd);
    else if (strncasecmp (cmd, "fcopy", 5) == 0)
      do_fcopy (cmd);    
    else if (strncasecmp (cmd, "ls", 2) == 0)
      do_readdir ();
    else if (strncasecmp (cmd, "cd", 2) == 0)
      do_cd (cmd);
    else if (strncasecmp (cmd, "diff", 4) == 0)
      do_diff (cmd);
    else if (strncmp (cmd, "QUIT", strlen ("QUIT")) == 0)
      break;
    else if (strncmp (cmd, "q", strlen ("q")) == 0)
      break;
    else {
      do_help ();
    }
  }
  sync_inodes ();
  release_bitmaps (&g_sb);
  brelse (g_sb.u.reiserfs_sb.s_sbh);
  check_and_free_mem ();
  return 0;
}

