/*
 * Copyright 1996, 1997, 1998 Hans Reiser
 */
#include <stdio.h>
#include <stdarg.h>
#include <asm/bitops.h>
#include <stdlib.h>
#include <string.h>
#include <mntent.h>

int test_and_set_bit (int nr, void * addr)
{
  int oldbit = test_bit (nr, addr);
  set_bit (nr, addr);
  return oldbit;
}

int test_and_clear_bit (int nr, void * addr)
{
  int oldbit = test_bit (nr, addr);
  clear_bit (nr, addr);
  return oldbit;
}


void die (char * fmt, ...)
{
  static char buf[1024];
  va_list args;

  va_start (args, fmt);
  vsprintf (buf, fmt, args);
  va_end (args);

  printf ("\n%s\n\n\n", buf);
  exit (-1);
}

static int g_allocated_memory = 0;
static int g_getmems = 0;
static int g_freemems = 0;

/*#define MEM_CONTROL*/

#ifdef MEM_CONTROL
/*&&&&&&&&&&&&&&&*/
int memnum = 0;
struct {
  char * buf;
  int len;
} mems [2000];
/*&&&&&&&&&&&&&&&*/

void add (char * buf, int len)
{
  int i;

  for (i = 0; i < memnum; i ++)
    if (mems[i].buf == 0) {
      mems[i].buf = buf;
      mems[i].len = len;
      return;
    }
  if (memnum == 2000)
    die ("add: no memory");
  mems[memnum].buf = buf;
  mems[memnum].len = len;
  memnum ++;
}

void del (char * buf)
{
  int i;

  for (i = 0; i < memnum; i ++)
    if (mems[i].buf == buf) {
      mems[i].buf = 0;
      return;
    }
  die ("del: buf not found");
}

int getlen (char * buf)
{
  int i;

  for (i = 0; i < memnum; i ++)
    if (mems[i].buf == buf) {
      return mems[i].len;
    }
  return -1;
}
#endif


void * getmem (int size)
{
  void * p;

  g_getmems ++;
  p = (void *)malloc (size);
  if (!p)
    die ("getmem: no more memory (%d)", size);

  g_allocated_memory += size;
  memset (p, 0, size);
/*&&&&&&&&&&&&&&*/
#ifdef MEM_CONTROL
  add (p, size);
#endif
/*&&&&&&&&&&&&&*/
  return p;
}


void * expandmem (void * p, int size, int by)
{
  p = realloc (p, size + by);
  if (!p)
    die ("expandmem: no more memory (%d)", size);

  memset (p + size, 0, by);
  g_allocated_memory += by;
  return p;
}


void freemem (void * p, int size)
{
/*&&&&&&&&&&&&&*/
#ifdef MEM_CONTROL
  if (size != getlen (p))
    die ("freemem: size %d, getlen %d", size, getlen (p));
  del (p);
#endif
/*&&&&&&&&&&&&&*/

  g_freemems ++;
  g_allocated_memory -= size;  
  free (p);
}


void check_and_free_buffer_mem (void);/* io.c */
void check_and_free_mem (void)
{
  check_and_free_buffer_mem ();
  
  if (g_allocated_memory != 0)
    die ("check_and_free_mem: non-freed memory %d (getmems %d, freemems %d)",
	 g_allocated_memory, g_getmems, g_freemems);
}


int is_mounted (char * device_name)
{
  FILE *f;
  struct mntent *mnt;

  if ((f = setmntent (MOUNTED, "r")) == NULL)
    return 0;

  while ((mnt = getmntent (f)) != NULL)
    if (strcmp (device_name, mnt->mnt_fsname) == 0)
      return 1;
  endmntent (f);

  return 0;
}






